import { Inject, Injectable } from '@angular/core';
import { HttpClient } from '@angular/common/http';
import { JwtHelperService } from '@auth0/angular-jwt';
import { Observable, timer } from 'rxjs';
import { map } from 'rxjs/operators';

import { AUTH_OPTIONS_TOKEN } from './auth.options.token';
import { StorageService } from '../storage/storage.service';
import { Events } from '../../providers/events';
import { AuthEvents } from './auth.events';
import { ResponseEntity } from './entities/response.entity';
import { UserEntity } from '../../entities/user.entity';
import { AuthDefaults } from './auth.defaults';
import { AuthConfig } from './entities/auth.config';

@Injectable({
  providedIn: 'root'
})
export class AuthService {

  private jwtHelper: JwtHelperService = new JwtHelperService();
  private isReadyForResolver: boolean = false;
  private _credentials: any = null;

  user: UserEntity;
  refreshSubscription: any;
  config: AuthConfig;

  get configuration(): AuthConfig {
    return this.config || {};
  }

  constructor(@Inject(AUTH_OPTIONS_TOKEN) config: AuthConfig,
              private http: HttpClient,
              private storage: StorageService,
              private events: Events) {

    this.config = config;
  }

  public isAuthenticated(): boolean {
    const token = this.storage.getRefreshToken();
    return token ? !this.jwtHelper.isTokenExpired(token) : false;
  }

  public login(credentials: any): Observable<boolean> {
    const deviceId = this.storage.getDeviceId();
    let params = new URLSearchParams();
    params.append('grant_type', AuthDefaults.grantTypes.login);
    params.append('username', credentials.username);
    params.append('password', credentials.password);

    this._credentials = credentials;

    return this.http.post(AuthDefaults.urls.simpleToken(this.config), params.toString(), { headers: AuthDefaults.headers(this.config, deviceId) })
      .pipe(map(response => this.processResponse(response as ResponseEntity)));

  }

  public loginFacebook(token: string): Observable<boolean> {
    const deviceId = this.storage.getDeviceId();
    let params = new URLSearchParams();
    params.append('grant_type', AuthDefaults.grantTypes.login);
    return this.http.post(AuthDefaults.urls.facebookToken(this.config, token), params.toString(), { headers: AuthDefaults.headers(this.config, deviceId) })
      .pipe(map(response => this.processResponse(response as ResponseEntity)));

  }

  public logout(emitEvent: boolean = true) {

    //localStorage.removeItem(myGlobals.ATT_OPTIONS_KEY);
    //this.optionsService.resetOptions();

    this.storage.resetStorage();
    this.unscheduleTokenRefresh();
    this.user = null;
    this.isReadyForResolver = false;

    if (emitEvent) {
      this.events.publish(AuthEvents.USER_LOGOUT);
    }

  }

  public ready(): Promise<boolean> {

    if (this.isReadyForResolver) {
      return Promise.resolve(this.isAuthenticated());
    }

    return Promise.resolve(this.isAuthenticated())
      .then(isAuthenticated => {

        // Check if there is an authenticated user
        if (isAuthenticated) {

          this.user = this.storage.getUser();
          this.events.publish(AuthEvents.USER_LOGIN);
          this.scheduleTokenRefresh();
          return isAuthenticated;

        }
        else {
          const refreshToken = this.storage.getRefreshToken();
          if (refreshToken) {
            return this.getNewToken();
          }
          return isAuthenticated;
        }

      })
      .then(isAuthenticated => {
        this.isReadyForResolver = true;
        return isAuthenticated;
      })
      .catch(err => {
        console.log(err);

        this.user = null;
        this.isReadyForResolver = true;
        return false;
      })

  }

  private getNewToken(): Promise<boolean> {
    const deviceId = this.storage.getDeviceId();
    const token = this.storage.getRefreshToken();
    let params = new URLSearchParams();
    let isRefreshResponse = true;
    if (!this.timeToRestartTokens()) {
      params.append('grant_type', AuthDefaults.grantTypes.token);
      params.append('refresh_token', token);
    } else {
      params.append('grant_type', AuthDefaults.grantTypes.login);
      params.append('username', this._credentials.username);
      params.append('password', this._credentials.password);
      isRefreshResponse = false;
    }
    console.log('getNewToken');

    return this.http
      .post(AuthDefaults.urls.simpleToken(this.config), params.toString(), { headers: AuthDefaults.headers(this.config, deviceId) })
      .pipe(map(res => this.processResponse(res as ResponseEntity, isRefreshResponse)))
      .toPromise()
      .then(result => Promise.resolve(result))
      .catch(err => {
        console.log(err);
        return Promise.resolve(false);
      });

  }

  private processResponse(response: ResponseEntity, isRefreshResponse?: boolean): boolean {

    if (!response.access_token || this.jwtHelper.isTokenExpired(response.access_token)) {
      throw new Error('Token Invalido');
    }

    this.storage.setAccessToken(response.access_token);
    this.storage.setRefreshToken(response.refresh_token);

    if (isRefreshResponse) {
      this.user = this.storage.getUser();
    } else {
      this.setTotalTimeRefresh();
      const claims = this.jwtHelper.decodeToken(response.access_token);
      this.user = new UserEntity({ user_name: claims.user_name });
      this.storage.setUser(this.user);
    }

    this.scheduleTokenRefresh();

    this.isReadyForResolver = true;
    console.info("lanzando evento USER_LOGIN")
    this.events.publish(AuthEvents.USER_LOGIN);
    return true;

  }

  private setTotalTimeRefresh() {
    // SET ONLY THE FIRST TIME
    const accessToken = this.storage.getAccessToken();
    const accessTokenTime = this.getAccessTokenRefreshDelay(accessToken);

    const refreshToken = this.storage.getRefreshToken();
    const refreshTokenTime = this.getAccessTokenRefreshDelay(refreshToken);

    this.storage.setRefreshTime(accessTokenTime);
    this.storage.setTotTime(accessTokenTime + refreshTokenTime);
  }

  private timeToRestartTokens(): boolean {
    this.storage.setTotTime(this.storage.getTotTime() - this.storage.getRefreshTime());
    const remaining = this.storage.getTotTime();

    return remaining - this.storage.getRefreshTime() < this.storage.getRefreshTime();
  }

  private scheduleTokenRefresh() {

    // Unsubscribe from previous refresh schedule
    this.unscheduleTokenRefresh();

    const accessToken = this.storage.getAccessToken();
    const delay = this.getAccessTokenRefreshDelay(accessToken);

  //  console.log ('New AccessToken: ' + accessToken + ', delay: ' + delay);

    this.refreshSubscription =
      timer(delay)
        .subscribe(() => {
          console.log('refreshSubscription subscriber');
          return this.getNewToken()
            .then(result => {
              if (!result) {
                console.info("lanzando evento USER_UNAUTHORIZED")
                this.events.publish(AuthEvents.USER_UNAUTHORIZED);
              } else {
                console.info("lanzando evento USER_REFRESH")
                this.events.publish(AuthEvents.USER_REFRESH);
              }
              return result;
            }).catch(err => console.warn("get new token ",err));
        }, err => {
          console.warn("Timer for token", err);
        });

  }

  private unscheduleTokenRefresh() {
    // Unsubscribe from the refresh
    if (this.refreshSubscription) {
      this.refreshSubscription.unsubscribe();
    }
  }

  private getAccessTokenRefreshDelay(token: string): number {
    let now: number = new Date().valueOf();
    let jwtExp: number = this.jwtHelper.decodeToken(token).exp;
    let exp: Date = new Date(0);

    console.log(jwtExp);
    exp.setUTCSeconds(jwtExp);

    console.log(exp.valueOf() - now);
    return exp.valueOf() - now;
  }

  public forceRefreshToken() {
    return this.getNewToken();
  }

}
